@@ -1,13 +1,10 @@
 #!/usr/bin/env python -u
 
 import datetime
-import platform
 import argparse
-import readline
 import logging
 import time
 import sys
-import os
 import re
 
 import wasatch
@@ -15,9 +12,10 @@ import wasatch
 from wasatch import utils
 from wasatch.WasatchBus         import WasatchBus
 from wasatch.WasatchDevice      import WasatchDevice
-from wasatch.BalanceAcquisition import BalanceAcquisition
+from wasatch.RealUSBDevice      import RealUSBDevice
+from wasatch.SpectrometerResponse import SpectrometerResponse
 
-VERSION = "2.2.6"
+VERSION = "2.3.0"
 
 log = logging.getLogger(__name__)
 
@@ -37,7 +35,7 @@ log = logging.getLogger(__name__)
 # @todo currently there is no support for scan averaging.  That is because 
 #       scan averaging is built into WasatchDevice.acquire_data, and configured
 #       via Feature
-class WasatchShell(object):
+class WasatchShell:
     
     def __init__(self):
         self.device = None                      # wasatch.WasatchDevice
@@ -53,66 +51,68 @@ class WasatchShell(object):
 
         self.configure_logging()
         self.input_tokens = None
+        self.dark_spectra = None
+        self.srm = False
 
         # pass-through calls to any of these gettors (note names are lowercased)
+        # SR = SpectrometerResponse
         self.gettors = {}
         for func_name in [ 
-            "get_actual_frames",
-            "get_actual_integration_time_us",
-            "get_ambient_temperature_degC",
-            "get_ccd_sensing_threshold",
-            "get_ccd_threshold_sensing_mode",
-            "get_dac",
-            "get_detector_gain",
-            "get_detector_gain_odd",
-            "get_detector_offset",
-            "get_detector_offset_odd",
-            "get_detector_tec_setpoint_degC",
-            "get_detector_tec_setpoint_raw",
-            "get_detector_temperature_degC",
-            "get_detector_temperature_raw",
-            "get_external_trigger_output",
-            "get_fan_enabled",
-            "get_fpga_firmware_version",
-            "get_high_gain_mode_enabled",
-            "get_integration_time_ms",
-            "get_lamp_enabled",
-            "get_laser_enabled",
-            "get_laser_interlock",
-            "get_laser_power_ramping_enabled",
-            "get_laser_temperature_degC",
-            "get_laser_temperature_raw",
-            "get_laser_watchdog_sec",
-            "get_microcontroller_firmware_version",
-            "get_mod_delay_us",
-            "get_mod_duration_us",
-            "get_mod_enabled",
-            "get_mod_linked_to_integration",
-            "get_mod_period_us",
-            "get_mod_width_us",
-            "get_opt_actual_integration_time",
-            "get_opt_area_scan",
-            "get_opt_cf_select",
-            "get_opt_data_header_tab",
-            "get_opt_has_laser",
-            "get_opt_horizontal_binning",
-            "get_opt_integration_time_resolution",
-            "get_opt_laser_control",
-            "get_raman_delay_ms",
-            "get_secondary_adc_calibrated",
-            "get_secondary_adc_raw",
-            "get_selected_adc",
-            "get_selected_laser",
-            "get_sensor_line_length",
-            "get_shutter_enabled",
-            "get_strobe_enabled",
-            "get_tec_enabled",
-            "get_trigger_delay",
-            "get_trigger_source",
-            "get_vr_continuous_ccd",
-            "get_vr_num_frames",
-            "has_laser_power_calibration",
-            "has_linearity_coeffs" ]:
+                "get_actual_frames",                        # SR
+                "get_actual_integration_time_us",           # SR
+                "get_ambient_temperature_degC",             # SR
+                "get_ccd_sensing_threshold",                # SR
+                "get_ccd_threshold_sensing_mode",           # SR
+                "get_dac",                                  # SR
+                "get_detector_gain",                        # SR
+                "get_detector_gain_odd",                    # SR
+                "get_detector_offset",                      # SR
+                "get_detector_offset_odd",                  # SR
+                "get_detector_tec_setpoint_degC",           # SR
+                "get_detector_tec_setpoint_raw",            # SR
+                "get_detector_temperature_degC",            # SR
+                "get_detector_temperature_raw",             # SR
+                "get_external_trigger_output",              # SR
+                "get_fan_enabled",                          # SR
+                "get_fpga_firmware_version",                # SR
+                "get_high_gain_mode_enabled",               # SR
+                "get_integration_time_ms",                  # SR
+                "get_lamp_enabled",                         # SR
+                "get_laser_enabled",                        # SR
+                "get_laser_power_attenuator",               # SR
+                "get_laser_interlock",                      # SR
+                "get_laser_temperature_degC",               # SR
+                "get_laser_temperature_raw",                # SR
+                "get_laser_watchdog_sec",                   # SR
+                "get_microcontroller_firmware_version",     # SR
+                "get_mod_delay_us",                         # SR
+                "get_mod_duration_us",                      # SR
+                "get_mod_enabled",                          # SR
+                "get_mod_linked_to_integration",            # SR
+                "get_mod_period_us",                        # SR
+                "get_mod_width_us",                         # SR
+                "get_opt_actual_integration_time",          # SR
+                "get_opt_area_scan",                        # SR
+                "get_opt_cf_select",                        # SR
+                "get_opt_data_header_tab",                  # SR
+                "get_opt_has_laser",                        # SR
+                "get_opt_horizontal_binning",               # SR
+                "get_opt_integration_time_resolution",      # SR
+                "get_opt_laser_control",                    # SR
+                "get_raman_delay_ms",                       # SR
+                "get_secondary_adc_calibrated",             # SR
+                "get_secondary_adc_raw",                    # SR
+                "get_selected_adc",                         # SR
+                "get_selected_laser",                       # SR
+                "get_sensor_line_length",                   # SR
+                "get_shutter_enabled",                      # SR
+                "get_strobe_enabled",                       # SR
+                "get_tec_enabled",                          # SR
+                "get_trigger_delay",                        # SR
+                "get_trigger_source",                       # SR
+                "get_vr_continuous_ccd",                    # SR
+                "get_vr_num_frames",                        # SR
+                "has_linearity_coeffs" ]:                   # BOOL
             self.gettors[func_name.lower()] = func_name
 
     # ##############################################################################
@@ -135,10 +135,8 @@ class WasatchShell(object):
         set_laser_enable                       - takes bool argument (on/off, true/false, 1/0)
         set_laser_power_mw                     - takes float argument
         set_laser_power_perc                   - takes int argument
-        set_laser_power_ramping_enable         - gradually ramp laser power in software
+        set_laser_power_attenuator             - takes uint8 argument
         set_laser_watchdog_sec                 - takes integer argument
-        set_acquisition_laser_trigger_enable   - takes bool argument
-        set_acquisition_laser_trigger_delay_ms - takes float argument
         set_tec_enable                         - takes bool argument
         set_detector_tec_setpoint_degc         - takes float argument
         set_detector_offset                    - override the "offset" added to pixel readings
@@ -162,6 +160,9 @@ class WasatchShell(object):
                                                     max_tries, x, unit [px, nm, cm]
                                                
         get_spectrum                           - print received spectrum
+        set_raman_intensity_correction_enable  - takes bool argument
+        get_dark                               - captures spectrum and stores as dark
+        clear_dark                             - clears a stored dark spectrum
         get_spectrum_pretty                    - graph received spectrum
         get_spectrum_save                      - save spectrum to filename as CSV
         get_config_json                        - return EEPROM as JSON string
@@ -267,6 +268,14 @@ class WasatchShell(object):
                         elif command == "get_spectrum":
                             self.get_spectrum(quiet=False)
 
+                        elif command == "get_dark":
+                            self.get_dark()
+                            self.display(1)
+
+                        elif command == "clear_dark":
+                            self.clear_dark()
+                            self.display(1)
+
                         elif command == "get_spectrum_pretty":
                             self.get_spectrum_pretty()
 
@@ -295,6 +304,9 @@ class WasatchShell(object):
                                                             end   = self.read_float(),
                                                             incr  = self.read_float())
 
+                        elif command == "set_raman_intensity_correction_enable":
+                            self.set_raman_intensity_correction_enable(self.device, self.read_bool())
+
                         elif command == "clear":
                             self.clear()
 
@@ -343,6 +355,10 @@ class WasatchShell(object):
                             self.device.change_setting("laser_power_perc", self.read_float())
                             self.display(1)
 
+                        elif command == "set_laser_power_attenuator":
+                            self.device.change_setting("laser_power_attenuator", self.read_int())
+                            self.display(1)
+
                         elif command == "set_laser_enable":
                             self.set_laser_enable(flag = self.read_bool())
 
@@ -358,10 +374,6 @@ class WasatchShell(object):
                             self.device.change_setting("detector_tec_setpoint_degC", self.read_float())
                             self.display(1)
 
-                        elif command == "set_laser_power_ramping_enable":
-                            self.device.change_setting("laser_power_ramping_enable", self.read_bool())
-                            self.display(1)
-
                         elif command == "set_detector_offset":
                             self.device.change_setting("detector_offset", self.read_int())
                             self.display(1)
@@ -370,14 +382,6 @@ class WasatchShell(object):
                             self.device.change_setting("scans_to_average", self.read_int())
                             self.display(1)
 
-                        elif command == "set_acquisition_laser_trigger_enable":
-                            self.device.change_setting("acquisition_laser_trigger_enable", self.read_bool())
-                            self.display(1)
-
-                        elif command == "set_acquisition_laser_trigger_delay_ms":
-                            self.device.change_setting("acquisition_laser_trigger_delay_ms", self.read_int())
-                            self.display(1)
-
                         elif command == "set_selected_laser":
                             self.device.change_setting("selected_laser", self.read_int())
                             self.display(1)
@@ -416,7 +420,7 @@ class WasatchShell(object):
 
                         else:
                             self.display("ERROR: unknown command: " + command)
-                    except Exception as ex:
+                    except:
                         log.critical("caught exception", exc_info=1)
                         log.info("disconnecting")
                         self.device.disconnect()
@@ -471,10 +475,11 @@ class WasatchShell(object):
 
         device_id = bus.device_ids[0]
         log.debug("open: trying to connect to %s", device_id)
+        device_id.device_type = RealUSBDevice(device_id)
         device = WasatchDevice(device_id)
 
         ok = device.connect()
-        if not ok: 
+        if not ok.data: 
             log.critical("open: can't connect to device on bus 1")
             return
 
@@ -502,7 +507,7 @@ class WasatchShell(object):
         # if laser has power calibration, require it and initialize accordingly 
         # (so if the user enables the laser, it won't fire at an out-of-bounds 
         # 100% unmodulated)
-        if self.device.hardware.has_laser_power_calibration():
+        if self.device.settings.eeprom.has_laser_power_calibration():
             log.info("laser has power calibration, so requiring modulation and initializing to max-rated power in mW")
             self.device.change_setting("laser_power_require_modulation", True)
             self.device.change_setting("laser_power_mW", self.device.settings.eeprom.max_laser_power_mW)
@@ -524,11 +529,44 @@ class WasatchShell(object):
     def run_gettor(self, command):
         func_name = self.gettors[command]
         value = getattr(self.device.hardware, func_name)()
+
+        if isinstance(value, SpectrometerResponse):
+            if value.error_msg:
+                self.display(f"ERROR: {value.error_msg}")
+                return
+            value = value.data
+
         if isinstance(value, bool):
             self.display(1 if value else 0)
         else:
             self.display(value)
 
+    def get_dark(self):
+        spectrum = self.get_spectrum(quiet=True)
+        self.dark_spectra = spectrum
+
+    def clear_dark(self):
+        self.dark_spectra = None
+
+    def set_raman_intensity_correction_enable(self, device: WasatchDevice, status: bool): # -> int 
+        if not device.settings.eeprom.has_raman_intensity_calibration():
+            self.display("Device has no Raman Intensity Calibration")
+            self.srm = False
+            return
+        else:
+            self.srm = status
+            self.display(1)
+            return
+
+    def srm_process(self, spectrum: list[float], device: WasatchDevice): # -> list[float] 
+
+        factors = device.settings.raman_intensity_factors
+        if factors is None or len(factors) != len(spectrum):
+            return 
+
+        spectrum = [px*fac for px, fac in zip(spectrum,factors)]
+        return spectrum
+
     ##
     # This calls WasatchDevice.acquire_data, rather than FID.get_line, because 
     # scan averaging, bad-pixel correction, acquisition laser trigger and other
@@ -539,13 +577,19 @@ class WasatchShell(object):
         self.device.change_setting("acquire", True, allow_immediate=False)
 
         # now collect the spectrum
-        reading = self.device.acquire_data()
+        reading_response = self.device.acquire_data()
+        reading = reading_response.data
         if reading is None or isinstance(reading, bool) or reading.spectrum is None:
             self.display("ERROR: get_spectrum failed")
             return
         spectrum = reading.spectrum
+        if self.dark_spectra is not None and len(self.dark_spectra) == len(spectrum):
+            spectrum = [spec-dark for spec, dark in zip(spectrum, self.dark_spectra)]
         log.debug("received %d pixels", len(spectrum))
 
+        if self.srm:
+            spectrum = self.srm_process(spectrum, self.device)
+
         if self.interpolated_x_axis_cm and self.device.settings.wavenumbers:
             spectrum = utils.interpolate_array(spectrum, 
                                                self.device.settings.wavenumbers, 
@@ -644,7 +688,7 @@ class WasatchShell(object):
         unit                    = "px"          if not self.has_input() else self.read_str()
 
         if not re.match('(px|cm|nm)$', unit):
-            return self.display("ERROR: invalid unit " + s)
+            return self.display(f"ERROR: invalid unit {unit}")
 
         pixel = None
         if x_value is not None:
